// Important APU defines
`include "APU.h"

`include "AUDIO_DAC.v"
`include "I2C_AV_Config.v"
`include "I2C_Controller.v"
//`include "Reset_Delay.v"
//`include "VGA_Audio_PLL.v"

// DDS units
`include "SyncRomSineTable.v"
`include "SyncRomSquareTable.v"
`include "SyncRomTriangleTable.v"
`include "SyncRomSawTable.v"

// Actual sampling frequency is 47.04 KHz

module APU
(
    iCLOCK_50, iCLOCK_27,
    iReset,
    // User Data
    iValid,
    iOp,
    iData,
    iAUDCtrlClk,
    iDLY_RST,
    // Audio Hookup
    ioAUD_ADCLRCK, iAUD_ADCDAT, ioAUD_DACLRCK,
    oAUD_DACDAT, ioAUD_BCLK, oAUD_XCK,
    //TD Reset
    oTD_RESET,
    // I2C protocol
    oI2C_SCLK, ioI2C_SDAT
);

// Clocks
input   wire            iCLOCK_50;
input   wire            iCLOCK_27;
input   wire            iReset;
input   wire            iAUDCtrlClk;
input   wire            iDLY_RST;

// User Control
input   wire            iValid;
input   wire    [15:0]  iOp;
input   wire    [31:0]  iData;

// Audio Hookup
inout   wire			ioAUD_ADCLRCK;
input	wire    		iAUD_ADCDAT;
inout	wire    		ioAUD_DACLRCK;
output	wire    		oAUD_DACDAT;
inout	wire    		ioAUD_BCLK;
output  wire		    oAUD_XCK;

// TD Reset
output  wire   			oTD_RESET;

// I2C Protocol
output  wire            oI2C_SCLK;
inout   wire            ioI2C_SDAT;


// Internal Nodes
wire wVGA_CTRL_CLK;
//wire wAUD_CTRL_CLK;
//wire wDLY_RST;

// Audio assigns
assign oTD_RESET = 1'b1;
assign ioAUD_ADCLRCK = ioAUD_DACLRCK;
assign oAUD_XCK = iAUDCtrlClk;

// 4 Wavetable Oscillators

// SINE
wire    [31:0]  wSineIncrement;
reg     [31:0]  rSineFrequency = 420;
reg     [31:0]  rSineAccumulator = 0;
wire    signed  [15:0]  wSineOut;
wire    [15:0]  wSineAmpOut;
reg             rSineOn = 0;
reg     [31:0]  rSineDuration = 32'hff;  // @50Mhz maxes out to be ~11.43 min
reg     [31:0]  rSineTime = 0;

// SQUARE
wire    [31:0]  wSquareIncrement;
reg     [31:0]  rSquareFrequency = 420;
reg     [31:0]  rSquareAccumulator = 0;
wire    signed  [15:0]  wSquareOut;
reg             rSquareOn = 0;
reg     [31:0]  rSquareDuration = 32'hff;  // @50Mhz maxes out to be ~11.43 min
reg     [31:0]  rSquareTime = 0;

// SAW
wire    [31:0]  wSawIncrement;
reg     [31:0]  rSawFrequency = 420;
reg     [31:0]  rSawAccumulator = 0;
wire    signed  [15:0]  wSawOut;
reg             rSawOn = 0;
reg     [31:0]  rSawDuration = 32'hff;  // @50Mhz maxes out to be ~11.43 min
reg     [31:0]  rSawTime = 0;

// TRIANGLE
wire    [31:0]  wTriangleIncrement;
reg     [31:0]  rTriangleFrequency = 420;
reg     [31:0]  rTriangleAccumulator = 0;
wire    signed  [15:0]  wTriangleOut;
reg             rTriangleOn = 0;
reg     [31:0]  rTriangleDuration = 32'hff;  // @50Mhz maxes out to be ~11.43 min
reg     [31:0]  rTriangleTime = 0;


// Noise
reg signed [31:0]  rRandomGen = 32'hAAAAAAAA;
reg     [17:0]  rNoiseCutoff = 18'h10000;   // start at .05
reg     [7:0]   rCoA = 1;
reg     [7:0]   rCoB = 1;
wire            wRandomBit;
reg             rNoiseOn = 1;
reg signed [15:0]  rNoise1;
reg signed [17:0]  rNoiseOut = 0;

reg     [31:0]  rNoiseDuration = 32'hff;  // @50Mhz maxes out to be ~11.43 min
reg     [31:0]  rNoiseTime = 0;

wire signed [17:0] wMultOut;


// Random noise generator, random bit will
// decide the shift value and the sign of the signal
// as well


assign wRandomBit = rRandomGen[31] ^ rRandomGen[30] ^ rRandomGen[29] ^ rRandomGen[9];

signed_mult
(
    wMultOut,
    rNoiseCutoff,
    (rNoiseOut - {wRandomBit, 17'h0})
);

/*always@(posedge iCLOCK_50)
begin
    if(rNoiseOn)
        begin
            if(rNoiseDuration == 0 || rNoiseTime <= rNoiseDuration)
                begin
                    rRandomGen <= {wRandomBit, rRandomGen[13:0], wRandomBit};
                    rNoiseOut <= {wRandomBit, 17'h0} + wMultOut;
                    if(rNoiseTime != rNoiseDuration)
                        begin
                            rNoiseTime = rNoiseTime + 1;
                        end
                    else if(rNoiseDuration != 0)
                        begin
                            rNoiseTime <= 0;
                            rNoiseOn <= 0;
                        end
                end
        end
    else
        begin
            rNoiseOut <= 0;
        end
end */

// Wave Table Increments
assign wSineIncrement = `DDSFactor * rSineFrequency;
assign wSquareIncrement = `DDSFactor * rSquareFrequency;
assign wSawIncrement = `DDSFactor * rSawFrequency;
assign wTriangleIncrement = `DDSFactor * rTriangleFrequency;

//always@(posedge iAUDCtrlClk)
always@(posedge ioAUD_DACLRCK)
begin
	if(iReset)
        begin
            rSineAccumulator <= 0;
            rSquareAccumulator <= 0;
            rSawAccumulator <= 0;
            rTriangleAccumulator <= 0;
        end
    else if(~iValid)
        begin
            // Reset some state?
        end
    else
        begin
            case(iOp[`action])
                // Wave off is similar to Note Off Convention
                `WaveOff:
                    begin
                        if(iOp[`target] == `Sine ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSineOn <= 0;
                                rSineTime <= 0;
                            end
                        if(iOp[`target] == `Square ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSquareOn <= 0;
                                rSquareTime <= 0;
                            end
                        if(iOp[`target] == `Saw ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSawOn <= 0;
                                rSawTime <= 0;
                            end
                        if(iOp[`target] == `Triangle ||
                           iOp[`target] == `AllWaves )
                            begin
                                rTriangleOn <= 0;
                                rTriangleTime <= 0;
                            end
                        if(iOp[`target] == `Noise ||
                           iOp[`target] == `AllWaves )
                            begin
                                rNoiseOn <= 0;
                                rNoiseTime <= 0;
                            end
                    end

                // Wave on is similar to Note On Convention
                `WaveOn:
                    begin
                        if(iOp[`target] == `Sine ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSineOn <= 1;
                                rSineTime <= 0;
                            end
                        if(iOp[`target] == `Square ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSquareOn <= 1;
                                rSquareTime <= 0;
                            end
                        if(iOp[`target] == `Saw ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSawOn <= 1;
                                rSawTime <= 0;
                            end
                        if(iOp[`target] == `Triangle ||
                           iOp[`target] == `AllWaves )
                            begin
                                rTriangleOn <= 1;
                                rTriangleTime <= 0;
                            end
                        if(iOp[`target] == `Noise ||
                           iOp[`target] == `AllWaves )
                            begin
                                rNoiseOn <= 1;
                                rNoiseTime <= 0;
                            end
                    end

                `SetWaveFrequency:
                    begin
                        if(iOp[`target] == `Sine ||
                           iOp[`target] == `AllWaves ) rSineFrequency <= iData;
                        if(iOp[`target] == `Square ||
                           iOp[`target] == `AllWaves ) rSquareFrequency <= iData;
                        if(iOp[`target] == `Saw ||
                           iOp[`target] == `AllWaves ) rSawFrequency <= iData;
                        if(iOp[`target] == `Triangle ||
                           iOp[`target] == `AllWaves ) rTriangleFrequency <= iData;
                        if(iOp[`target] == `Noise ||
                           iOp[`target] == `AllWaves ) rNoiseCutoff <= iData[17:0];
                    end

                `SetWaveDuration:
                    begin
                        if(iOp[`target] == `Sine ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSineDuration <= iData;
                                rSineTime <= 0;
                            end
                        if(iOp[`target] == `Square ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSquareDuration <= iData;
                                rSquareTime <= 0;
                            end
                        if(iOp[`target] == `Saw ||
                           iOp[`target] == `AllWaves )
                            begin
                                rSawDuration <= iData;
                                rSawTime <= 0;
                            end
                        if(iOp[`target] == `Triangle ||
                           iOp[`target] == `AllWaves )
                            begin
                                rTriangleDuration <= iData;
                                rTriangleTime <= 0;
                            end
                        if(iOp[`target] == `Noise ||
                           iOp[`target] == `AllWaves )
                            begin
                                rNoiseDuration <= iData;
                                rNoiseTime <= 0;
                            end
                    end

            endcase
        end

    // DDS Accumulator Increment
    if(rSineOn)
        begin
            if(rSineDuration == 0 || rSineTime <= rSineDuration)
                begin
                    rSineAccumulator <= rSineAccumulator + wSineIncrement;
                    if(rSineTime != rSineDuration)
                        begin
                            rSineTime <= rSineTime + 1;
                        end
                    else if(rSineDuration != 0)
                        begin
                            rSineTime <= 0;
                            rSineOn <= 0;
                        end
                end
        end
    else
        begin
            rSineAccumulator <= 0;
            rSineTime <= 0;
        end
    if(rSquareOn)
        begin
            if(rSquareDuration == 0 || rSquareTime <= rSquareDuration)
                begin
                    rSquareAccumulator <= rSquareAccumulator + wSquareIncrement;
                    if(rSquareTime != rSquareDuration)
                        begin
                            rSquareTime <= rSquareTime + 1;
                        end
                    else if(rSquareDuration != 0)
                        begin
                            rSquareTime <= 0;
                            rSquareOn <= 0;
                        end
                end
        end
    else
        begin
            rSquareAccumulator <= 0;
            rSquareTime <= 0;
        end
    if(rSawOn)
        begin
            if(rSawDuration == 0 || rSawTime <= rSawDuration)
                begin
                    rSawAccumulator <= rSawAccumulator + wSawIncrement;
                    if(rSawTime != rSawDuration)
                        begin
                            rSawTime <= rSawTime + 1;
                        end
                    else if(rSawDuration != 0)
                        begin
                            rSawTime <= 0;
                            rSawOn <= 0;
                        end
                end
        end
    else
        begin
            rSawAccumulator <= 0;
            rSawTime <= 0;
        end
    if(rTriangleOn)
        begin
            if(rTriangleDuration == 0 || rTriangleTime <= rTriangleDuration)
                begin
                    rTriangleAccumulator <= rTriangleAccumulator + wTriangleIncrement;
                    if(rTriangleTime != rTriangleDuration)
                        begin
                            rTriangleTime <= rTriangleTime + 1;
                        end
                    else if(rTriangleDuration != 0)
                        begin
                            rTriangleTime <= 0;
                            rTriangleOn <= 0;
                        end
                end
        end
    else
        begin
            rTriangleAccumulator <= 0;
            rTriangleTime <= 0;
        end
    if(rNoiseOn)
        begin
            if(rNoiseDuration == 0 || rNoiseTime <= rNoiseDuration)
                begin
                    rRandomGen <= {rRandomGen[30:0], wRandomBit};
                    rNoiseOut <= {wRandomBit, 17'h0} + wMultOut;
                    if(rNoiseTime != rNoiseDuration)
                        begin
                            rNoiseTime <= rNoiseTime + 1;
                        end
                    else if(rNoiseDuration != 0)
                        begin
                            rNoiseTime <= 0;
                            rNoiseOn <= 0;
                        end
                end
        end
    else
        begin
            rNoiseOut <= 0;
            rNoiseTime <= 0;
        end
end

wire    signed  [17:0]   wWaveOut;

assign wWaveOut = (wSineOut + wSquareOut + wSawOut + wTriangleOut + rNoiseOut[17:2]);

/*assign wSineAmpOut = // Attack
                     (rSineTime <= rSineAttack) ? wSineOut * (rSineTime * (32'hffffffff / rSineAttack)) :
                     // Decay
                     (rSineTime > rSineAttack &&
                      rSineTime <= rSineAttack + rSineDecay) ? wSineOut :
                     // Sustain
                     (rSineTime > rSineAttack + rSineDecay &&
                      rSineTime < rSineDuration) ? wSineOut * (32'hffffffff / rSineSustain):
                     // Release
                     wSineOut * (32'hffffffff / rSineSustain);
                     */

// DDS Modules
// Sine
SyncRomSineTable syncromoSinetable
(
	.iCLK		(iCLOCK_50),
	.iAddress	(rSineAccumulator[31:24]),
	.oSine		(wSineOut)
);

// Square
SyncRomSquareTable syncromosquaretable
(
	.iCLK		(iCLOCK_50),
	.iAddress	(rSquareAccumulator[31:24]),
	.oSquare	(wSquareOut)
);

// Saw
SyncRomSawTable syncromosawtable
(
	.iCLK		(iCLOCK_50),
	.iAddress	(rSawAccumulator[31:24]),
	.oSaw		(wSawOut)
);

// Triangle
SyncRomTriangleTable syncromotriangletable
(
	.iCLK		(iCLOCK_50),
	.iAddress	(rTriangleAccumulator[31:24]),
	.oTriangle	(wTriangleOut)
);


/*
Reset_Delay r5
(
	.iCLK	(iCLOCK_50),
	.oRESET (wDLY_RST)
);


VGA_Audio_PLL p3
(
	.areset		(~wDLY_RST),
	.inclk0		(iCLOCK_27),
	.c0			(wVGA_CTRL_CLK),
	.c1			(wAUD_CTRL_CLK),
	.c2			(VGA_CLK)
);
*/

I2C_AV_Config u3
(
	// Host Side
	.iCLK		(iCLOCK_50),
	.iRST_N		(~iReset),
	// I2C Side
	.I2C_SCLK	(oI2C_SCLK),
	.I2C_SDAT	(ioI2C_SDAT)
);

AUDIO_DAC u4
(
	// Audio Side
	.oAUD_BCK		(ioAUD_BCLK),
	.oAUD_DATA		(oAUD_DACDAT),
	.oAUD_LRCK		(ioAUD_DACLRCK),

	// Inputs
	.iAUD_extR		(wWaveOut[15:0]),
	.iAUD_extL		(wWaveOut[15:0]),

	// Control Signals
	.iSrc_Select	(1'b1),
	.iCLK_18_4		(iAUDCtrlClk),
	.iRST_N			(iDLY_RST)
);

endmodule

module signed_mult (out, a, b);
	output 		[17:0]	out;
	input 	signed	[17:0] 	a;
	input 	signed	[17:0] 	b;
	wire	signed	[17:0]	out;
	wire 	signed	[35:0]	mult_out;
	assign mult_out = a * b;
	assign out = {mult_out[35], mult_out[32:16]};
endmodule
