// This is the network routing layer,  it basically acts transparently to
// the processor taking in an input message and dispatching or setting a
// flag to the processor that a message is waiting in the stack

`include "ndmaref.h"

module NLayer
(
    CLK, MRST,

    // ID
    idFlag, id, oSelfID,

    // Receive
    rxFlag, rxData,

    // Send
    dxFlag, dest_id, send_data,
    // Network Connect
    nBusIn, sBusIn, eBusIn, wBusIn,
    nBusOut, sBusOut, eBusOut, wBusOut
);

input   wire            MRST;
input   wire            CLK;
// ID
input   wire            idFlag;
input   wire    [7:0]   id;
output  wire    [7:0]   oSelfID;

// Recieve
output  wire            rxFlag;
output  wire    [7:0]   rxData;

// if we want to send a message
// assert the dxFlag when all the fields are valid
input   wire            dxFlag;
reg                     dxFlag_r;

input   wire    [7:0]   dest_id;
input   wire    [7:0]   send_data;



// We have 4 directions (this is technically scalable)
/*****************************************************
* Message is 32 bits:
*  8 bit Destination ID
*  8 bit Message
*  8 bit Origination ID
*  4 bit Age
*  2 bit origin dispatch direction
*  2 bit last taken direction
******************************************************/
input   wire    [31:0]  nBusIn;
input   wire    [31:0]  sBusIn;
input   wire    [31:0]  eBusIn;
input   wire    [31:0]  wBusIn;

output  wire    [31:0]  nBusOut;
output  wire    [31:0]  sBusOut;
output  wire    [31:0]  eBusOut;
output  wire    [31:0]  wBusOut;

// This obtains the value if the connect is "on"
// if not the output busses are assigned to the wire values.
reg	    [31:0]  nBus_R;
reg     [31:0]  sBus_R;
reg     [31:0]  eBus_R;
reg     [31:0]  wBus_R;

// Adaptive Network Counters
reg     [7:0]   n_cnt;
reg     [7:0]   s_cnt;
reg     [7:0]   e_cnt;
reg     [7:0]   w_cnt;

// Adaptive Network Control
wire            n_Thru;
wire            s_Thru;
wire            w_Thru;
wire            e_Thru;

// Receieve
reg             rxnFlag;
reg             rxsFlag;
reg             rxeFlag;
reg             rxwFlag;

reg     [7:0]   rxnData;
reg     [7:0]   rxsData;
reg     [7:0]   rxeData;
reg     [7:0]   rxwData;

// This is our 8 bit IDs
reg     [7:0]   self_id;

assign oSelfID = self_id;

// Adaptive Logic
assign n_Thru = (n_cnt > `OptVal);
assign s_Thru = (s_cnt > `OptVal);
assign w_Thru = (w_cnt > `OptVal);
assign e_Thru = (e_cnt > `OptVal);


//assign n_Thru = 0;
//assign s_Thru = 0;
//assign w_Thru = 0;
//assign e_Thru = 0;

// Register Thru Debugging
always @(n_Thru or s_Thru or w_Thru or e_Thru) begin
	if(n_Thru) $display("* North thru on id: %d cut", self_id);
	if(s_Thru) $display("* South thru on id: %d cut", self_id);
	if(e_Thru) $display("* East thru on id: %d cut", self_id);
	if(w_Thru) $display("* West thru on id: %d cut", self_id);
end


// this is our thru mux where the connection
// is enabled based on the x_Thru control signal
assign nBusOut = (s_Thru) ? sBusIn : nBus_R;
assign sBusOut = (n_Thru) ? nBusIn : sBus_R;
assign eBusOut = (w_Thru) ? wBusIn : eBus_R;
assign wBusOut = (e_Thru) ? eBusIn : wBus_R;

assign rxFlag = ( rxnFlag | rxsFlag | rxeFlag | rxwFlag );

assign rxData = (rxnFlag) ? rxnData :
                (rxsFlag) ? rxsData :
                (rxeFlag) ? rxeData :
                (rxwFlag) ? rxwData : 8'hzz;

always @(posedge CLK) begin
	if(MRST) begin
		n_cnt <= 0;
		s_cnt <= 0;
		e_cnt <= 0;
		w_cnt <= 0;
		rxnFlag <= 0;
		rxsFlag <= 0;
		rxeFlag <= 0;
		rxwFlag <= 0;
        rxnData <= 0;
        rxsData <= 0;
        rxeData <= 0;
        rxwData <= 0;
		dxFlag_r <= 0;
		//self_id = 1;
	end

	if(dxFlag) begin
		if(!dxFlag_r) begin
            $display("dxflag did:%d sd:%d sid:%d", dest_id, send_data, self_id);
			nBus_R <= {dest_id, send_data, self_id, 4'b0, `north, `north};
			sBus_R <= {dest_id, send_data, self_id, 4'b0, `south, `south};
			eBus_R <= {dest_id, send_data, self_id, 4'b0, `east, `east};
			wBus_R <= {dest_id, send_data, self_id, 4'b0, `west, `west};
		end else begin
			nBus_R <= 0;
			sBus_R <= 0;
			eBus_R <= 0;
			wBus_R <= 0;
		end
		dxFlag_r <= 1;
	end else begin
		dxFlag_r <= 0;
		nBus_R <= 0;
		sBus_R <= 0;
		eBus_R <= 0;
		wBus_R <= 0;
	end
end

always @(idFlag) begin
	if(idFlag) begin
		self_id <= id;
	end
end

// The way message passing occurs on this net is wave-like
// So if a message comes in from the horizontal directions it is propogated in all directions
// except for the one in which it came.  If it incomes from the vertical direction it continues
// to propogate in that direction.

always @(nBusIn) begin
//always @(negedge CLK) begin
	if(nBusIn[`dest_id] == self_id && !n_Thru)
	    begin
	    	$display("MR_N d_id:%d o_id:%d msg:%d", nBusIn[`dest_id], nBusIn[`origin_id], nBusIn[`msg_data]);
	    	$stop;
	    	rxnFlag <= 1;
            rxnData <= nBusIn[`msg_data];
	    	if(n_cnt != 0)
	    	    begin
	    	        n_cnt = n_cnt - 1;
                end
	    end
	else
		begin
		    rxnFlag <= 0;
            rxnData <= 0;
		    if(nBusIn != 0)
		    	begin
		    	    if(n_cnt < `OptVal) n_cnt = n_cnt + 1;
		    	    /*$display("Pass @%d msg N->S dest: %d", self_id, nBusIn[`dest_id]);
		    	    $display("From id %d", nBusIn[`origin_id]);
		    	    $display("Data: %d", nBusIn[`msg_data]);*/
		    	    case(nBusIn[`o_dir])
		    	    	// this should never happen
		    	    	`north:	$display("north bus error! n_id: %d", self_id);
		    	    	// this means we pass it on to the south bus
		    	    	// regardless of last direction
		    	    	`south:	sBus_R <= `delay nBusIn;
		    	    	// this means we pass it on to the south bus
		    	    	// regardless of last direction
		    	    	`east:	sBus_R <= `delay nBusIn;
		    	    	// this means we pass it on to the south bus
		    	    	// regardless of last direction
		    	    	`west:	sBus_R <= `delay nBusIn;
		    	    endcase
		    	end
		    else
		        begin
		        	sBus_R <= 0;
		        	eBus_R <= 0;
		        	wBus_R <= 0;
		        end
		end
end

always @(sBusIn) begin
	if(sBusIn[`dest_id] == self_id && !s_Thru)
	    begin
	    	$display("MR_S d_id:%d o_id:%d msg:%d", sBusIn[`dest_id], sBusIn[`origin_id], sBusIn[`msg_data]);
	    	$stop;
	    	rxsFlag <= 1;
            rxsData <= sBusIn[`msg_data];
	    	if(s_cnt != 0)
	    	    begin
	    	        s_cnt = s_cnt - 1;
                end
	    end
	else
	    begin
	    	rxsFlag <= 0;
            rxsData <= 0;
	    	if(sBusIn != 0)
	    	    begin
	    	    	s_cnt = s_cnt + 1;
	    	    	/*$display("Pass @%d msg S->N dest: %d", self_id, sBusIn[`dest_id]);
	    	    	$display("From id %d", sBusIn[`origin_id]);
	    	    	$display("Data: %d", sBusIn[`msg_data]);*/
	    	    	case(sBusIn[`o_dir])
	    	    			// this should never happen
	    	    			`south:	$display("south bus error! n_id: %d", self_id);
	    	    			// this means we pass it on to the south bus
	    	    			// regardless of last direction
	    	    			`north:	nBus_R <= `delay sBusIn;
	    	    			// this means we pass it on to the south bus
	    	    			// regardless of last direction
	    	    			`east:	nBus_R <= `delay sBusIn;
	    	    			// this means we pass it on to the south bus
	    	    			// regardless of last direction
	    	    			`west:	nBus_R <= `delay sBusIn;
	    	    	endcase
	    	    end
	    	else
	    	    begin
	    	    	nBus_R <= 0;
	    	    	eBus_R <= 0;
	    	    	wBus_R <= 0;
	    	    end
	    end
end

// East / West Behaviour
always @(eBusIn) begin
	if(eBusIn[`dest_id] == self_id && !e_Thru)
	    begin
	    	$display("MR_E d_id:%d o_id:%d msg:%d", eBusIn[`dest_id], eBusIn[`origin_id], eBusIn[`msg_data]);
	    	$stop;
	    	rxeFlag <= 1;
            rxeData <= eBusIn[`msg_data];
	    	if(e_cnt != 0)
	    	    begin
	    	        e_cnt = e_cnt - 1;
                end
	    end
	else
	    begin
	    	rxeFlag <= 0;
            rxeData <= 0;
	    	if(eBusIn != 0)
	    	    begin
	    	    	e_cnt = e_cnt + 1;
	    	    	/*$display("Pass @%d msg E-> dest: %d", self_id, eBusIn[`dest_id]);
	    	    	$display("From id %d", eBusIn[`origin_id]);
	    	    	$display("Data: %d", eBusIn[`msg_data]);*/
	    	    	case(eBusIn[`o_dir])
	    	    			// this should never happen
	    	    			`north:	$display("east bus error! n_id: %d", self_id);
	    	    			// this should never happen
	    	    			`south:	$display("east bus error! n_id: %d", self_id);
	    	    			// this should never happen
	    	    			`east:	$display("east bus error! n_id: %d", self_id);
	    	    			// We pass it on in all directions (except east)
	    	    			`west:
	    	    			    begin
	    	    					sBus_R <= `delay eBusIn;
	    	    					nBus_R <= `delay eBusIn;
	    	    					wBus_R <= `delay eBusIn;
	    	    				end
	    	    	endcase
	    	    end
	    	else
	    	    begin
	    	    	nBus_R <= 0;
	    	    	sBus_R <= 0;
	    	    	wBus_R <= 0;
	    	    end
	    end
end

always @(wBusIn) begin
	if(wBusIn[`dest_id] == self_id && !w_Thru)
	    begin
	    	$display("MR_W d_id:%d o_id:%d msg:%d", wBusIn[`dest_id], wBusIn[`origin_id], wBusIn[`msg_data]);
	    	$stop;
	    	rxwFlag <= 1;
            rxwData <= wBusIn[`msg_data];
	    	//$finish;
	    	if(w_cnt != 0)
	    	    begin
	    	        w_cnt = w_cnt - 1;
                end
	    end
	else
	    begin
	    	rxwFlag <= 0;
            rxwData <= 0;
	    	if(wBusIn != 0)
	    	    begin
	    	    	w_cnt = w_cnt + 1;
	    	    	$display("Pass @%d msg W-> dest: %d", self_id, wBusIn[`dest_id]);
	    	    	$display("From id %d", wBusIn[`origin_id]);
	    	    	$display("Data: %d", wBusIn[`msg_data]);
	    	    	case(wBusIn[`o_dir])
	    	    			// this should never happen
	    	    			`north:	$display("west bus error! n_id: %d", self_id);
	    	    			// this should never happen
	    	    			`south:	$display("west bus error! n_id: %d", self_id);
	    	    			// this should never happen
	    	    			`west:	$display("west bus error! n_id: %d", self_id);
	    	    			// We pass it on in all directions (except west)
	    	    			`east:
	    	    			    begin
	    	    					sBus_R <= `delay wBusIn;
	    	    					nBus_R <= `delay wBusIn;
	    	    					eBus_R <= `delay wBusIn;
	    	    				end
	    	    	endcase
	    	    end
	    	else
	    	    begin
	    	    	nBus_R <= 0;
	    	    	sBus_R <= 0;
	    	    	eBus_R <= 0;
	    	    end
	    end
end

endmodule
