module RS232BootLoader
(
    iCLK, iMRST,
    // Transmit
    oASCII,         // Transmit ASCII Value
    oValid_TX,      // Valid Transmit
    iComplete_TX,   // Transmit Complete
    // Recieve
    iASCII,         // Receive ASCII Value
    iBufferEmpty,   // Buffer Empty Value
    oValid_RX,      // Valid Receive
    iComplete_RX,    // Receive Complete
    oState,
    // Boot Loader Hookup
    oBootLoad,
    oPC,
    oIns,
    oWE_N
);

input   wire            iCLK;
input   wire            iMRST;

// Transmit
output  reg     [7:0]   oASCII;
output  reg             oValid_TX = 0;
input   wire            iComplete_TX;

// Receive
input   wire    [7:0]   iASCII;
input   wire            iBufferEmpty;
input   wire            iComplete_RX;
output  reg             oValid_RX = 0;

// Boot Loader Hook Up
output  reg             oBootLoad = 0;
output  wire    [31:0]  oPC;
output  wire    [31:0]  oIns;
output  reg             oWE_N = 1;

assign oPC = rPC;
assign oIns = rIns;

output  wire    [15:0]  oState;
assign oState = rState;

// Internal States
reg     [15:0]  rState = init;
reg     [7:0]   rASCII;

reg     [3:0]   rPC_cnt = 0;
reg     [31:0]  rPC = 0;
reg     [31:0]  rPC_out = 0;

reg     [3:0]   rIns_cnt = 0;
reg     [31:0]  rIns = 0;
reg     [31:0]  rIns_out = 0;


parameter   init = 0,
            // Boot Receive
            boot_pend = 1,
            boot_b = 2, boot_b_rx = 3,
            boot_bo = 4, boot_bo_rx = 5,
            boot_boo = 6, boot_boo_rx = 7,
            boot_boot = 8, boot_boot_rx = 9,
            // Ready Reply
            reply_r = 10, reply_r_tx = 11,
            reply_re = 12, reply_re_tx = 13,
            reply_rea = 14, reply_rea_tx = 15,
            reply_read = 16, reply_read_tx = 17,
            reply_ready = 18, reply_ready_tx = 19,
            // Get PC
            gPC = 20, gPC_rx = 21,
            // Respond PC
            dPC = 22, dPC_tx = 23,
            // Get Instruction
            gIns = 24, gIns_rx = 25,
            // Respond Instruction
            dIns = 26, dIns_tx = 27,
            // Write To Mem
            mem_set = 28, mem_done = 39,
            // Done
            Done = 30;

always@(posedge iCLK) begin
    if(iMRST)
        begin
            rState <= init;
            oValid_TX <= 0;
            oValid_RX <= 0;
            rPC_cnt <= 0;
            rPC <= 0;
            rIns_cnt <= 0;
            rIns <= 0;
            oWE_N <= 1;
            oBootLoad <= 0;
        end
    else
        begin
            case(rState)
                init:
                    begin
                        // Wait for input of "boot"
                        // "boot" == 0x62, 0x6F, 0x6F, 0x74
                        if(!iBufferEmpty && !iComplete_RX)
                            begin
                                oValid_RX <= 1;
                                rState <= boot_b_rx;
                            end
                    end

                boot_b_rx:
                    begin
                        if(iComplete_RX)
                            begin
                                oValid_RX <= 0;
                                if(iASCII == 8'h62)
                                    begin
                                        rASCII <= iASCII;
                                        rState <= boot_bo;
                                    end
                                else
                                    begin
                                        rState <= init;
                                    end
                            end
                    end

                boot_bo:
                    begin
                        if(!iBufferEmpty && !iComplete_RX)
                            begin
                                oValid_RX <= 1;
                                rState <= boot_bo_rx;
                            end
                    end

                boot_bo_rx:
                    begin
                        if(iComplete_RX)
                            begin
                                oValid_RX <= 0;
                                if(iASCII == 8'h6F)
                                    begin
                                        rASCII <= iASCII;
                                        rState <= boot_boo;
                                    end
                                else
                                    begin
                                        rState <= init;
                                    end
                            end
                    end

                boot_boo:
                    begin
                        if(!iBufferEmpty && !iComplete_RX)
                            begin
                                oValid_RX <= 1;
                                rState <= boot_boo_rx;
                            end
                    end

                boot_boo_rx:
                    begin
                        if(iComplete_RX)
                            begin
                                oValid_RX <= 0;
                                if(iASCII == 8'h6F)
                                    begin
                                        rASCII <= iASCII;
                                        rState <= boot_boot;
                                    end
                                else
                                    begin
                                        rState <= init;
                                    end
                            end
                    end

                boot_boot:
                    begin
                        if(!iBufferEmpty && !iComplete_RX)
                            begin
                                oValid_RX <= 1;
                                rState <= boot_boot_rx;
                            end
                    end

                boot_boot_rx:
                    begin
                        if(iComplete_RX)
                            begin
                                oValid_RX <= 0;
                                if(iASCII == 8'h74)
                                    begin
                                        rASCII <= iASCII;
                                        oBootLoad <= 1;
                                        rState <= reply_r;
                                    end
                                else
                                    begin
                                        rState <= init;
                                    end
                            end
                    end

                // Reply with "ready"
                // "ready" == 0x72, 0x65, 0x61, 0x64, 0x79
                reply_r:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= 8'h72;
                                oValid_TX <= 1;
                                rState <= reply_r_tx;
                            end
                    end

                reply_r_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                rState <= reply_re;
                            end
                    end

                reply_re:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= 8'h65;
                                oValid_TX <= 1;
                                rState <= reply_re_tx;
                            end
                    end

                reply_re_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                rState <= reply_rea;
                            end
                    end

                reply_rea:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= 8'h61;
                                oValid_TX <= 1;
                                rState <= reply_rea_tx;
                            end
                    end

                reply_rea_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                rState <= reply_read;
                            end
                    end

                reply_read:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= 8'h64;
                                oValid_TX <= 1;
                                rState <= reply_read_tx;
                            end
                    end

                reply_read_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                rState <= reply_ready;
                            end
                    end

                reply_ready:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= 8'h79;
                                oValid_TX <= 1;
                                rState <= reply_ready_tx;
                            end
                    end

                reply_ready_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                rState <= gPC;
                            end
                    end

                gPC:
                    begin
                        if(!iBufferEmpty && !iComplete_RX)
                            begin
                                oValid_RX <= 1;
                                rState <= gPC_rx;
                            end
                    end

                gPC_rx:
                    begin
                        if(iComplete_RX)
                            begin
                                oValid_RX <= 0;
                                rASCII <= iASCII;
                                rPC <= {iASCII, rPC[31:8]};
                                if(rPC_cnt < 3)
                                    begin
                                        rPC_cnt <= rPC_cnt + 1;
                                        rState <= gPC;
                                    end
                                else
                                    begin
                                        rPC_cnt <= 0;
                                        rPC_out <= {iASCII, rPC[31:8]};
                                        rState <= dPC;
                                    end
                            end
                    end

                dPC:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= rPC_out[7:0];
                                oValid_TX <= 1;
                                rState <= dPC_tx;
                            end
                    end

                dPC_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                if(rPC_cnt < 3)
                                    begin
                                        rPC_cnt <= rPC_cnt + 1;
                                        rPC_out <= {8'h00, rPC_out[31:8]};
                                        rState <= dPC;
                                    end
                                else
                                    begin
                                        if(rPC == 32'hFFFFFFFF)
                                            begin
                                                // Should not write to
                                                // PC = 0xFFFFFFFF
                                                // (this designates)
                                                // transmission over
                                                rState <= Done;
                                            end
                                        else
                                            begin
                                                rState <= gIns;
                                            end
                                    end
                            end
                    end

                gIns:
                    begin
                        if(!iBufferEmpty && !iComplete_RX)
                            begin
                                oValid_RX <= 1;
                                rState <= gIns_rx;
                            end
                    end

                gIns_rx:
                    begin
                        if(iComplete_RX)
                            begin
                                oValid_RX <= 0;
                                rASCII <= iASCII;
                                rIns <= {iASCII, rIns[31:8]};
                                if(rIns_cnt < 3)
                                    begin
                                        rIns_cnt <= rIns_cnt + 1;
                                        rState <= gIns;
                                    end
                                else
                                    begin
                                        rIns_cnt <= 0;
                                        rIns_out <= {iASCII, rIns[31:8]};
                                        rState <= dIns;
                                    end
                            end
                    end

                dIns:
                    begin
                        if(!iComplete_TX)
                            begin
                                oASCII <= rIns_out[7:0];
                                oValid_TX <= 1;
                                rState <= dIns_tx;
                            end
                    end

                dIns_tx:
                    begin
                        if(iComplete_TX)
                            begin
                                oValid_TX <= 0;
                                if(rIns_cnt < 3)
                                    begin
                                        rIns_cnt <= rIns_cnt + 1;
                                        rIns_out <= {8'h00, rIns_out[31:8]};
                                        rState <= dIns;
                                    end
                                else
                                    begin
                                        rState <= mem_set;                                    end
                                    end
                    end

                mem_set:
                    begin
                        if(rPC_cnt == 3 && rIns_cnt == 3)
                            begin
                                oWE_N <= 0;
                                rState <= mem_done;
                            end
                    end

                mem_done:
                    begin
                        oWE_N <= 1;
                        rPC_cnt <= 0;
                        rIns_cnt <= 0;
                        //rIns <= 0;
                        //rPC <= 0;
                        rState <= gPC;
                    end

                Done:
                    begin
                        oWE_N <= 1;
                        rPC_cnt <= 0;
                        rIns_cnt <= 0;
                        oBootLoad <= 0;
                        rState <= init;
                    end
            endcase
        end
end



endmodule
