//`include "../../IO/HexDecode/HexDecode.v"
`include "../../IO/PS2/ScanCodeDecoder.v"
`include "../../IO/PS2/PS2.v"
//`include "../../MEM/M4K/M4KMEM.v"

// This is a buffer that hooks up into the PS2 port and whenever a key
// is pressed it will save that ASCII value to memory and then increment
// the write position.  The buffer is circular and allows up to 256
// characters to be written before it starts to lose data.

// The read convention is such that when the iValid line goes high the
// memory is read from the "read" position and then the position is
// incremented.  Once oComplete goes high the value in oASCIIOut
// will be valid and can be read. Only one read is allowed for every
// valid high.

module PS2InputBuffer
(
    iCLK, iMRST,
    iPS2_CLK, iPS2_DAT,
    iValid,
    oBufferEmpty,
    oASCIIOut,
    oComplete
);

// Control Lines
input   wire            iCLK;
input   wire            iMRST;

// PS2 Port
input   wire            iPS2_CLK;
input   wire            iPS2_DAT;

// Inputs
input   wire            iValid;

// Outputs
output  wire            oBufferEmpty;
output  reg     [7:0]   oASCIIOut;
output  reg             oComplete;

wire    [7:0]   wScancode;
wire    [7:0]   wASCII;
reg     [7:0]   rASCII;

wire    [7:0]   wReadData;

reg     [7:0]   rMemWritePos = 0;
reg     [7:0]   rMemReadPos = 0;
wire    [7:0]   wMemAddr;

assign wMemAddr = (iValid) ? rMemReadPos : rMemWritePos;

assign oBufferEmpty = (rMemWritePos == rMemReadPos);

wire            wShiftEn;
wire            wKeyValid;
reg             rKeyRead = 0;
reg     [4:0]   rcKeyRead = 0;
reg             rWriteKey_N = 1;

ScanCodeDecoder scd1(
	.iScancode(wScancode),
	.oCharOut (wASCII),
	.iShiftEn (wShiftEn)
);

PS2 ps2(
    .iPS2_CLK(iPS2_CLK), .iPS2_DAT(iPS2_DAT),
    .iMRST(iMRST),
    .oKeyValid(wKeyValid),
    .oScancode(wScancode),
    .oShiftEnable(wShiftEn)
);

M4KMEM InputBuffer(
    .iCLK(iCLK), .iMRST(iMRST),
    .iWE_N(rWriteKey_N), .iAddr(wMemAddr),
    .iWriteData(rASCII), .oReadData(wReadData)
);
// m4kmem parameters
defparam InputBuffer.pRAM_ADDR_WIDTH = 8,
         InputBuffer.pRAM_UNIT_COUNT = 256,
         InputBuffer.pRAM_UNIT_WIDTH = 8;
// end InputBuffer parameters

reg [4:0]   rState = 0;
    reg [4:0]   rState2 = 0;

always@(posedge iCLK) begin
     if(~iValid)
        begin
            oComplete <= 0;
            rState2 <= 0;
        end
     else if(!oBufferEmpty)
        begin
            if(rState2 == 0)
                begin
                    // Wait one cycle to let the
                    // memory adjust to the new
                    // memory address
                    rState2 <= 1;
                end
            else if(rState2 == 1)
                begin
                    oASCIIOut <= wReadData;
                    rMemReadPos <= rMemReadPos + 1;
                    oComplete <= 1;
                    rState2 <= 2;
                end
            /*else if(rState2 == 2)
                begin
                    oComplete <= 1;
                    rState2 <= 3;
                end    */

        end

    if(~wKeyValid)
        begin
            rWriteKey_N <= 1;
            rState <= 0;
        end
    else
        begin
            case(rState)
                0:
                    begin
                        rState <= 1;
                    end
                1:
                    begin
                        rASCII <= wASCII;
                        rState <= 2;
                    end
                2:
                    begin
                        rWriteKey_N <= 0;
                        rState <= 3;
                    end
                3:
                    begin
                        rWriteKey_N <= 1;
                        rState <= 4;
                    end

                4:
                    begin
                        rMemWritePos <= rMemWritePos + 1;
                        rState <= 5;
                    end

            endcase
        end
end


endmodule
