//`include "../../MEM/M4K/M4KMEM.v"

// This RS-232 module is a simple module which interfaces through the
// RS-232 port.  This could potentially be used to boot-load a CPU or
// let the CPU communicate through the RS-232 to the computer.

`define UART_INTERFACE_DE2 .oUART_TXD(UART_TXD), .iUART_RXD(UART_RXD)

module RS232InputBuffer
(
    iCLK, iMRST,
    // RS-232 Connections
    oUART_TXD,  // Transmit
    iUART_RXD,  // Receive
    // Buffer Interface
    // Transmit
    iASCII_TX,
    iValid_TX,
    oComplete_TX,

    // Receive
    iValid,
    oBufferEmpty,
    oASCIIOut,
    oComplete
);

input   wire            iCLK;
input   wire            iMRST;

// RS-232 Connect
output  wire            oUART_TXD;
input   wire            iUART_RXD;

// Buffer Interface
input   wire    [7:0]   iASCII_TX;
input   wire            iValid_TX;
output  reg             oComplete_TX = 0;

input   wire            iValid;
output  wire            oBufferEmpty;
output  reg     [7:0]   oASCIIOut;
output  reg             oComplete;


reg     [7:0]   rASCII;
reg             rWriteASCII_N = 1;

// Input Buffer Stuff
wire    [7:0]   wReadData;
reg     [7:0]   rMemWritePos = 0;
reg     [7:0]   rMemReadPos = 0;
wire    [7:0]   wMemAddr;

assign wMemAddr = (iValid) ? rMemReadPos : rMemWritePos;
assign oBufferEmpty = (rMemWritePos == rMemReadPos);

// Baud Generator
// ************************************************************
parameter ClkFreq = 50000000;
parameter Baud = 115200;
parameter Baud8 = 8*Baud;
parameter BaudGenAccWidth = 16;
parameter BaudGenInc = ((Baud<<(BaudGenAccWidth-4))+(ClkFreq>>5))/(ClkFreq>>4);
wire [BaudGenAccWidth:0] Baud8GenInc = ((Baud8<<(BaudGenAccWidth-7))+(ClkFreq>>8))/(ClkFreq>>7);

reg     [BaudGenAccWidth:0] BaudGenAcc = 0;
reg     [BaudGenAccWidth:0] Baud8GenAcc = 0;

always@(posedge iCLK) begin
    BaudGenAcc <= BaudGenAcc[BaudGenAccWidth-1:0] + BaudGenInc;
    Baud8GenAcc <= Baud8GenAcc[BaudGenAccWidth-1:0] + Baud8GenInc;
end

wire BaudTick = BaudGenAcc[BaudGenAccWidth];
wire Baud8Tick = Baud8GenAcc[BaudGenAccWidth];

/*reg BaudTick = 0;
always@(posedge BaudGenAcc[BaudGenAccWidth-1]) begin
    BaudTick <= ~BaudTick;
end

reg Baud8Tick = 0;
always@(posedge Baud8GenAcc[BaudGenAccWidth-1]) begin
    Baud8Tick <= ~Baud8Tick;
end
*/


// ************************************************************

// Transmit
// ************************************************************
reg     [3:0]   state_TX;
reg             MUX_BIT_TX;
always@(posedge BaudTick) begin
    if(~iValid_TX || iMRST)
        begin
            state_TX <= 0;
            oComplete_TX <= 0;
        end
    else
        begin
            case(state_TX)
                4'b0000:    if(~oComplete_TX) state_TX <= 4'b0100;
                4'b0100:    state_TX <= 4'b1000;  // Start Bit
                4'b1000:    state_TX <= 4'b1001;  // Bit 0
                4'b1001:    state_TX <= 4'b1010;  // Bit 1
                4'b1010:    state_TX <= 4'b1011;  // Bit 2
                4'b1011:    state_TX <= 4'b1100;  // Bit 3
                4'b1100:    state_TX <= 4'b1101;  // Bit 4
                4'b1101:    state_TX <= 4'b1110;  // Bit 5
                4'b1110:    state_TX <= 4'b1111;  // Bit 6
                4'b1111:    state_TX <= 4'b0001;  // Bit 7
                4'b0001:    state_TX <= 4'b0010;  // Stop Bit 1
                4'b0010:
                    begin
                        state_TX <= 4'b0011;  // Stop Bit 2
                        //oComplete_TX <= 1;
                    end
                4'b0011:
                    begin
                        oComplete_TX <= 1;
                        state_TX <= 4'b0011;
                    end

                //default: state_TX <= 4'b0000;
            endcase
        end
end

always@(state_TX[2:0]) begin
    case(state_TX[2:0])
        0:  MUX_BIT_TX <= iASCII_TX[0];
        1:  MUX_BIT_TX <= iASCII_TX[1];
        2:  MUX_BIT_TX <= iASCII_TX[2];
        3:  MUX_BIT_TX <= iASCII_TX[3];
        4:  MUX_BIT_TX <= iASCII_TX[4];
        5:  MUX_BIT_TX <= iASCII_TX[5];
        6:  MUX_BIT_TX <= iASCII_TX[6];
        7:  MUX_BIT_TX <= iASCII_TX[7];
    endcase
end

wire txd = (state_TX < 4) | (state_TX[3] & MUX_BIT_TX);

assign oUART_TXD = txd;
// ************************************************************

// Receive
// ************************************************************
reg     [3:0]   state_RX;
reg     [1:0]   sync_RX;
reg     [1:0]   cnt_RX;
reg             bit_RX;
reg     [2:0]   bit_space = 0;
reg     [7:0]   data_RX;

reg     [4:0]   bit_cnt = 0;

// Filter Data
always@(posedge iCLK) begin
    if(iUART_RXD && cnt_RX != 2'b11)
        begin
            cnt_RX <= cnt_RX + 1;
        end
    else if(~iUART_RXD && cnt_RX != 2'b00)
        begin
            cnt_RX <= cnt_RX - 1;
        end

    if(cnt_RX == 2'b00)
        begin
            bit_RX <= 0;

        end
    else if(cnt_RX == 2'b11)
        begin
            bit_RX <= 1;
        end
end

always@(posedge Baud8Tick) begin
    if(state_RX == 0 || iMRST)
        begin
            bit_space <= 0;
        end
    else
        begin
            bit_space <= bit_space + 1;
            next_bit <= (bit_space == 3'b111);
        end
end

reg next_bit;// = (bit_space == 3'b111);

// Receive State Machine
always@(posedge Baud8Tick) begin
    if(iMRST)
        begin
            state_RX <= 0;
        end
    else
        begin
            case(state_RX)
                4'b0000:
                    begin
                        if(~bit_RX)
                            begin
                                bit_cnt <= 0;
                                state_RX <= 4'b1000;   // Start bit
                            end
                    end
                4'b1000:
                    begin
                        if(next_bit)
                            begin
                                data_RX <= {bit_RX, data_RX[7:1]};
                                if(bit_cnt < 7)
                                    begin
                                        bit_cnt <= bit_cnt + 1;
                                    end
                                else
                                    begin
                                        state_RX <= 4'b0001;
                                    end
                            end

                    end

                4'b0001:
                    begin
                        //if(data_RX != 0)    // Ignore NULL Characters
                        if(next_bit)
                            begin
                                rWriteASCII_N <= 0;
                                rASCII <= data_RX;
                                state_RX <= 4'b0010;
                            end
                    end

               4'b0010:
                   begin
                       rWriteASCII_N <= 1;
                       if(next_bit)     // Ignore NULL Characters
                            begin
                                rMemWritePos <= rMemWritePos + 1;
                                state_RX <= 4'b0000;
                            end
                   end
            endcase
        end
end
// ************************************************************

// Input Buffer State Machine
// ************************************************************
reg [4:0]   rState2 = 0;

always@(posedge iCLK) begin
     if(~iValid || iMRST)
        begin
            oComplete <= 0;
            rState2 <= 0;
        end
     else if(!oBufferEmpty)
        begin
            if(rState2 == 0)
                begin
                    // Wait one cycle to let the
                    // memory adjust to the new
                    // memory address
                    rState2 <= 1;
                end
            else if(rState2 == 1)
                begin
                    oASCIIOut <= wReadData;
                    rMemReadPos <= rMemReadPos + 1;
                    oComplete <= 1;
                    rState2 <= 2;
                end
        end
end
// *************************************************************

M4KMEM InputBuffer(
    .iCLK(iCLK), .iMRST(iMRST),
    .iWE_N(rWriteASCII_N), .iAddr(wMemAddr),
    .iWriteData(rASCII), .oReadData(wReadData)
);
// m4kmem parameters
defparam InputBuffer.pRAM_ADDR_WIDTH = 8,
         InputBuffer.pRAM_UNIT_COUNT = 256,
         InputBuffer.pRAM_UNIT_WIDTH = 8;
// end InputBuffer parameters



endmodule
