`include "SRAM.v"

`define SCREEN_WIDTH 640
`define SCREEN_HEIGHT 480
`define SCREEN_MARGIN 7

// Creates a 640x480 screen buffer
// addressable through x, y coordinates 10 bits wide
// Each memory element contains 2 pixels, each pixel is 8 bits deep
// this means that each color contains 2 bits, or a total of 63 total
// colors, with 2 bits for alpha channel if ever needed.

// This module will write to screen the RGB value to screen while
// iWE_N is high. It needs two cycles to do so, so if X,Y,R,G, or B
// changes in between these cycles the plot will not register.
// WE_N must go high again before next cycle to ensure this occurs

// This module can be easily altered to compensate for any resolution
// with a pixel of any bit depth since as long as the convention of
// plotting one pixel at a time is maintained you could cut up the buffer
// however you please

// Implementation Template:
// module SRAM_SCREEN_BUFFER
// (
//     // User Data
//     .iCoordX( ), .iCoordY( ),
//     .iWE_N( ),
//     .iWriteR( ), .iWriteG( ), .iWriteB( ),
//     .oReadData( ),
//     // Actual SRAM
//     `SRAM_INTERFACE_DE2
// );
// `SRAM_INTERFACE_DE2 defined in SRAM.v in ./

module SRAM_SCREEN_BUFFER
(
    iCLK, iClear,
    // User Data
    iCoordX, iCoordY,
    iWE_N,
    iWriteR, iWriteG, iWriteB,
    oReadData,
    // Actual SRAM
    ioSRAM_DQ, oSRAM_ADDR,
    oSRAM_UB_N, oSRAM_LB_N,
    oSRAM_WE_N,
    oSRAM_CE_N, oSRAM_OE_N
);

input wire iCLK;
input wire iClear;  // This will simply set writeDQ to 0
// User Data
input   wire [9:0]  iCoordX;
input   wire [9:0]  iCoordY;
input   wire        iWE_N;
input   wire [1:0]  iWriteR;
input   wire [1:0]  iWriteG;
input   wire [1:0]  iWriteB;
output  wire [7:0]  oReadData;
// Actual SRAM
inout	wire [15:0]	ioSRAM_DQ;		//	SRAM Data bus 16 Bits
output	wire [17:0]	oSRAM_ADDR;		//	SRAM Address bus 18 Bits
output  wire	    oSRAM_UB_N;		//	SRAM High-byte Data Mask
output	wire	    oSRAM_LB_N;		//	SRAM Low-byte Data Mask
output	wire	    oSRAM_WE_N;		//	SRAM Write Enable
output	wire	    oSRAM_CE_N;		//	SRAM Chip Enable
output	wire	    oSRAM_OE_N;		//	SRAM Output Enable

wire [17:0] wAddr;
wire [7:0] wDataWrite;

reg rWE_N = 1;
reg rWE1_N = 1;

//reg rRE = 0;
reg [15:0] rWriteDQ;
reg [15:0] rTempDQ;

reg [9:0] rX, rY;
reg [9:0] rX1, rY1;

assign wAddr = (rWE1_N) ? {iCoordY[8:0], iCoordX[9:1]} :
                         {rY1[8:0], rX1[9:1]};

assign wDataWrite = {iWriteR, iWriteG, iWriteB, 2'b00};

// SRAM Stuff
assign ioSRAM_DQ = (rWE1_N) ? 16'hzzzz : rWriteDQ;

//reg [3:0] rState = 0;

assign oSRAM_ADDR = wAddr;
assign oSRAM_UB_N = 0;	            // hi byte select enabled
assign oSRAM_LB_N = 0;		        // lo byte select enabled
assign oSRAM_CE_N = 0;		        // chip is enabled
assign oSRAM_WE_N = rWE1_N;	// Notice WE is negated
assign oSRAM_OE_N = 0;		        //output enable is overidden by WE

// If we're not writing then the Read Data is set to the first most
// significant 8 bits of DQ or least significant based on the LSB of
// iXCoord.
assign oReadData = (rWE1_N) ?
                        ((~iCoordX[0]) ? ioSRAM_DQ[15:8] : ioSRAM_DQ[7:0])
                                                       : 8'hzzzz;
always@(posedge iCLK) begin
    if(iClear)
        begin
            rWE1_N <= 0;
            rWriteDQ <= 0;

            if(rX1 == 640)
                begin
                    rX1 <= 0;
                    rY1 <= rY1 + 10'd1;
                end
            else rX1 <= rX1 + 10'd1;
        end
    else
        begin
            // Stage 1
            rTempDQ <= ioSRAM_DQ;
            rX <= iCoordX;
            rY <= iCoordY;
            rWE_N <= iWE_N;

            // Stage 2
            if(~rX[0]) rWriteDQ <= {wDataWrite, rTempDQ[7:0]};
            else rWriteDQ <= {rTempDQ[15:8], wDataWrite};
            rX1 <= rX;
            rY1 <= rY;
            rWE1_N <= rWE_N;
        end

    /*if(iWE_N == 1 && rState != 1)
        begin
            rState <= 0;
            rWE_N <= 1;
            //rTempDQ <= ioSRAM_DQ;
        end
    else begin
        case(rState)
            0:
                begin
                    rTempDQ <= ioSRAM_DQ;
                    rX <= iCoordX;
                    rY <= iCoordY;
                    rWE_N <= 1;
                    rState <= 1;
                end
            1:
                begin
                    if(~iCoordX[0])
                        begin
                            rWriteDQ <= {wDataWrite, rTempDQ[7:0]};
                            //rTempDQ <= {wDataWrite, rTempDQ[7:0]};
                        end
                    else
                        begin
                            rWriteDQ <= {rTempDQ[15:8], wDataWrite};
                            //rTempDQ <= {rTempDQ[15:8], wDataWrite};
                        end
                    rWE_N <= 0;
                    rState <= 2;
                end
        endcase
    end*/
end

endmodule
