// This module takes input as two pairs of (x, y) coordinates
// and a valid signal.  The signal is used to write to a hooked
// in memory (generally an SRAM module).

// We base the assumption that the resolution cannot be larger than
// 640 x 480.  This means that a 10 bit wide field for x, y should be
// more than enough for a while.

// Currently using 8 bpp so iColor and oWriteData will be 8 bits wide
// Based on writing to SRAM so addressing is based on a 16 bit deep
// address.

// The convention is LineValid goes high, then remains high until
// LineComplete goes high, and then LineValid can go low and
// LineComplete will go low again.  The caller must wait for LineComplete
// to finish before changing the inputs otherwise there will be no
// change in the output.

module Brensenham(
    iLock, iCLK,
    // User Inputs
    iValidLine,
    iX1, iY1,
    iX2, iY2,
    iColor,
    // Output to signify completion
    oLineComplete,
    // Screen Buffer Hookup
    oXCoord, oYCoord,
    oWE_N,
    oWriteR, oWriteG, oWriteB
);

input   wire            iLock;  // Halt's plotting
input   wire            iCLK;
input   wire            iValidLine;
input   wire    [9:0]   iX1;
input   wire    [9:0]   iY1;
input   wire    [9:0]   iX2;
input   wire    [9:0]   iY2;
input   wire    [7:0]   iColor;

output  reg     [9:0]   oXCoord;
output  reg     [9:0]   oYCoord;
output  reg             oWE_N = 1'b1;
output  wire    [1:0]   oWriteR, oWriteG, oWriteB;
output  reg             oLineComplete = 0;

// State Variables
reg     [3:0]   rState = Init;
reg             rSwap;
reg     [9:0]   rXCoord;
reg     [9:0]   rYCoord;
reg     [9:0]   rX1, rX2, rY1, rY2;
reg     [9:0]   rDeltaX, rDeltaY;
integer         rError;     // needs to be signed
reg             rStepY;
reg             rLock;

assign oWriteR = iColor[7:6];
assign oWriteG = iColor[5:4];
assign oWriteB = iColor[3:2];

parameter Init = 4'd0,
          Init1 = 4'd1,
          Init2 = 4'd2,
          Init3 = 4'd3,
          Init4 = 4'd4,
          Walk = 4'd5,
          Walk2 = 4'd6,
          Walk3 = 4'd7,
          DoneState = 4'd8;

//always@(iValidLine or iX1 or iX2 or iY1 or iY2 or rState) begin
always@(posedge iCLK) begin
    if(~iValidLine)
        begin
            rState <= Init;
            oWE_N <= 1'b1;
            oLineComplete <= 0;
        end
    else
    if(iLock)
        begin
            case(rState)
                //boolean steep := abs(y1 - y0) > abs(x1 - x0)
                //  if steep then
                //    swap(x0, y0)
                //    swap(x1, y1)
                Init:
                    begin
                        oLineComplete <= 0;
                        // Rather than pipeline this we just make it
                        // extremely conditional
                        if((iY1 > iY2 && iX1 > iX2) && ((iY1 - iY2) > (iX1 - iX2)) ||
                           (iY2 > iY1 && iX1 > iX2) && ((iY2 - iY1) > (iX1 - iX2)) ||
                           (iY1 > iY2 && iX2 > iX1) && ((iY1 - iY2) > (iX2 - iX1)) ||
                           (iY2 > iY1 && iX2 > iX1) && ((iY2 - iY1) > (iX2 - iX1))   )
                            begin
                                rX1 <= iY1;
                                rY1 <= iX1;
                                rX2 <= iY2;
                                rY2 <= iX2;
                                rSwap <= 1;
                            end
                        else
                            begin
                                rX1 <= iX1;
                                rY1 <= iY1;
                                rX2 <= iX2;
                                rY2 <= iY2;
                                rSwap <= 0;
                            end

                       rState <= Init2;
                    end
                //  if x0 > x1 then
                //    swap(x0, x1)
                //    swap(y0, y1)
                Init2:
                    begin
                        if(rX1 > rX2)
                            begin
                                rX1 <= rX2;
                                rX2 <= rX1;
                                rY1 <= rY2;
                                rY2 <= rY1;
                            end
                        rState <= Init3;
                    end

                //  int deltax := x1 - x0
                //    int deltay := abs(y1 - y0)
                //    int error := -deltax / 2
                //    int ystep
                //    int y := y0
                //    if y0 < y1 then ystep := 1 else ystep := -1
                //    for x from x0 to x1
                Init3:
                    begin
                        rDeltaX <= rX2 - rX1;

                        if(rY2 > rY1)
                            begin
                                rDeltaY <= rY2 - rY1;
                                rStepY <= 1;
                            end
                        else
                            begin
                                rDeltaY <= rY1 - rY2;
                                rStepY <= 0;
                            end

                        rState <= Init4;
                    end

                Init4:
                    begin
                        rError <= rDeltaY - rDeltaX;

                        rXCoord <= rX1;
                        rYCoord <= rY1;

                        rLock <= 1;
                        rState <= Walk;
                    end


                //  if steep then plot(y,x) else plot(x,y)
                //    error := error + deltay
                //    if error > 0 then
                //    y := y + ystep
                //    error := error - deltax
                Walk:
                    begin
                        if(rLock) begin
                            if(rSwap)
                                begin
                                    // plot (y, x)
                                    oXCoord <= rYCoord;
                                    oYCoord <= rXCoord;
                                    oWE_N <= 1'b0;
                                end
                            else
                                begin
                                    // plot (x, y)
                                    oXCoord <= rXCoord;
                                    oYCoord <= rYCoord;
                                    oWE_N <= 1'b0;
                                end

                            rError <= rError + rDeltaY;
                            rState <= Walk2;
                        end else
                            begin
                                // If lock is 0 then we reset it
                                // and try again!
                                rState <= Walk;
                                rLock <= 1;
                            end


                    end
                 Walk2:
                    begin
                        oWE_N <= 1'b1;
                        rLock <= 1;

                        if(rError >= 0)
                            begin
                                if(rStepY) rYCoord <= rYCoord + 10'd1;
                                else rYCoord <= rYCoord - 10'd1;
                                rError <= rError - rDeltaX;
                            end

                        if(rXCoord == rX2)
                             begin
                                 rState <= DoneState;
                             end
                        else
                             begin
                                 rXCoord <= rXCoord + 1;
                                 rState <= Walk3;
                             end
                    end

                 Walk3:
                    begin
                        rState <= Walk;
                        rLock <= 1;
                        oWE_N <= 1'b1;
                    end

                 DoneState:
                    begin
                        // The line is done
                        // ValidLine must go low before another line
                        // can be drawn
                        oWE_N <= 1'b1;
                        oLineComplete <= 1;
                    end
             endcase
        end else
            begin
                rLock <= 0;
                oWE_N <= 1'b1;
            end
end


endmodule
