// This module takes input as two pairs of (x, y) coordinates
// and a valid signal.  The signal is used to write to a hooked
// in memory (generally an SRAM module).

// We base the assumption that the resolution cannot be larger than
// 640 x 480.  This means that a 10 bit wide field for x, y should be
// more than enough for a while.

// Currently using 8 bpp so iColor and oWriteData will be 8 bits wide
// Based on writing to SRAM so addressing is based on a 16 bit deep
// address.

module Brensenham(
    iLock, iCLK,
    // User Inputs
    iValidLine,
    iX1, iY1,
    iX2, iY2,
    iColor,
    // Screen Buffer Hookup
    oXCoord, oYCoord,
    oWE_N,
    oWriteR, oWriteG, oWriteB
);

input   wire            iLock;  // Halt's plotting
input   wire            iCLK;
input   wire            iValidLine;
input   wire    [9:0]   iX1;
input   wire    [9:0]   iY1;
input   wire    [9:0]   iX2;
input   wire    [9:0]   iY2;
input   wire    [7:0]   iColor;

output  reg     [9:0]   oXCoord;
output  reg     [9:0]   oYCoord;
output  reg             oWE_N = 1'b1;
output  wire    [1:0]   oWriteR, oWriteG, oWriteB;

// State Variables
reg     [3:0]   rState = 0;

reg             rSwap;
reg     [9:0]   rXCoord;
reg     [9:0]   rYCoord;

reg     [9:0]   rX1, rX2, rY1, rY2;
reg     [9:0]   rDeltaX, rDeltaY;
reg     [9:0]   rError;
reg     [9:0]   rStepY;

reg             rLock;

assign oWriteR = 2'b11;//iColor[7:6];
assign oWriteG = 2'b11;//iColor[5:4];
assign oWriteB = 2'b11;//iColor[3:2];

parameter Init = 4'd0,
          Init1 = 4'd1,
          Init2 = 4'd2,
          Init3 = 3'd3,
          Walk = 4'd4,
          DoneState = 4'd5;

//always@(iValidLine or iX1 or iX2 or iY1 or iY2 or rState) begin
always@(posedge iCLK) begin
    if(~iValidLine)
        begin
            rState <= Init;
            oWE_N <= 1'b1;
        end
    else
    if(iLock)
        begin
            case(rState)
                //boolean steep := abs(y1 - y0) > abs(x1 - x0)
                //  if steep then
                //    swap(x0, y0)
                //    swap(x1, y1)
                Init:
                    begin
                        // Rather than pipeline this we just make it
                        // extremely conditional
                        if((iY1 > iY2 && iX1 > iX2) && ((iY1 - iY2) > (iX1 - iX2)) ||
                           (iY1 < iY2 && iX1 > iX2) && ((iY2 - iY1) > (iX1 - iX2)) ||
                           (iY1 > iY2 && iX1 < iX2) && ((iY1 - iY2) > (iX2 - iX1)) ||
                           (iY1 < iY2 && iX1 < iX2) && ((iY2 - iY1) > (iX2 - iX1))   )
                            begin
                                rX1 <= iY1;
                                rY1 <= iX1;
                                rX2 <= iY2;
                                rY2 <= iX1;
                                rSwap <= 1;
                            end
                        else
                            begin
                                rX1 <= iX1;
                                rY1 <= iY1;
                                rX2 <= iX2;
                                rY2 <= iY2;
                                rSwap <= 0;
                            end

                        if(iValidLine) rState <= Init2;
                    end
                //  if x0 > x1 then
                //    swap(x0, x1)
                //    swap(y0, y1)
                Init2:
                    begin
                        if(rX1 > rX2)
                            begin
                                rX1 <= rX2;
                                rX2 <= rX1;
                                rY1 <= rY2;
                                rY2 <= rY1;
                            end
                        rState <= Init3;
                    end

                //  int deltax := x1 - x0
                //    int deltay := abs(y1 - y0)
                //    int error := -deltax / 2
                //    int ystep
                //    int y := y0
                //    if y0 < y1 then ystep := 1 else ystep := -1
                //    for x from x0 to x1
                Init3:
                    begin
                        rDeltaX <= rX2 - rX1;
                        if(rY2 > rY1) rDeltaY <= rY2 - rY1;
                        else rDeltaY <= rY1 - rY2;
                        // x1 and x2 flipped to avoid a conversion to 2c
                        rError <= (rY2 - rY1) - (rX2 - rX1);

                        if(rY1 < rY2) rStepY <= 1;
                        else rStepY <= -1;

                        rYCoord <= rY1;
                        rXCoord <= rX1;
                        rState <= Walk;
                    end
                //  if steep then plot(y,x) else plot(x,y)
                //    error := error + deltay
                //    if error > 0 then
                //    y := y + ystep
                //    error := error - deltax
                Walk:
                    begin
                        rLock <= 1;
                        //if(rLock) begin
                            if(rSwap)
                                begin
                                    // plot (y, x)
                                    oXCoord <= rYCoord;
                                    oYCoord <= rXCoord;
                                    oWE_N <= 1'b0;
                                end
                            else
                                begin
                                    // plot (x, y)
                                    oXCoord <= rXCoord;
                                    oYCoord <= rYCoord;
                                    oWE_N <= 1'b0;
                                end
                            if(rError + rDeltaY > 0)
                                begin
                                    rYCoord <= rYCoord + rStepY;
                                    rError <= rError + rDeltaY - rDeltaX;
                                end
                            else rError <= rError + rDeltaY;
                            if(rXCoord == rX2)
                                begin
                                    rState <= DoneState;
                                end
                            else
                                begin
                                    rXCoord <= rXCoord + 1;
                                end
                        //end else oWE_N <= 1'b1;
                    end

                 DoneState:
                    begin
                        // The line is done
                        // ValidLine must go low before another line
                        // can be drawn
                        oWE_N <= 1'b1;
                    end

                 /*0:
                    begin
                        oXCoord <= iX1;
                        oYCoord <= iY1;
                        oWE_N <= 1'b0;
                        rState <= 1;
                    end
                1:
                    begin
                        rLock <= 1;
                        rState <= 2;
                        oWE_N <= 1'b1;
                    end
                2:
                    begin
                         if(rLock)
                            begin
                                oWE_N <= 1'b0;
                                oXCoord <= oXCoord + 1;
                                oYCoord <= oYCoord + 1;
                            end
                        else oWE_N <= 1'b1;
                        rState <= 3;
                    end
                3:
                    begin
                        oWE_N <= 1'b1;
                        //if(oXCoord < 100 && oYCoord < 100) rState <= 1;
                        //else
                        rState <= 4;
                    end

                4:
                    begin
                        if(oXCoord < 320) rState <= 1;
                    end
                    */
            endcase
        end else
            begin
                rLock <= 0;
                oWE_N <= 1'b1;
            end
end


endmodule
