// The GPU is an interface for the screen buffer, VGA controller,
// and some hardware acceleration features such as ASCII display and
// bresenham line algorithm.

// The GPU takes in instructions in a 16 bit field format.
// Each OP code consists of a 6 bit op code and a 10 bit data
// field.   In the case of setting Color or Char the top 2 bits of
// the 10 bit data field in the instruction is not used.

// When the caller signals Valid the GPU will start to perform the
// appropriate instruction.  When the instruction is complete then
// the GPU will signal the oComplete line.  Then the caller must
// lower the valid line before inititating a new instruction.

// See GPU.h for GPU op code defines

`include "../../MEM/SRAM/SRAM_SCREEN_BUFFER.v"
`include "../VGACNT/VGACNT.v"
`include "../Bresenham/Bresenham.v"
`include "../ASCIIVGA/ASCIIVGA.v"
`include "../PixelPlot/PixelPlot.v"

`include  "GPU.h"

module GPU
(
    // Clocks
    iCLK50, iCLK27,
    // User Data
    iReset,
    iValid,
    iIns,
    oComplete,
    oVGA_CTRL_CLK,
    oAUDCtrlClk,
    oDLY_RST,
    // VGA Interface
    oVGA_CLK,
    oVGA_HS, oVGA_VS,
    oVGA_BLANK, oVGA_SYNC,
    oVGA_R, oVGA_G, oVGA_B,
    // TD Reset
    oTD_RESET,
    // SRAM Interface
    ioSRAM_DQ, oSRAM_ADDR,
    oSRAM_UB_N, oSRAM_LB_N,
    oSRAM_WE_N,
    oSRAM_CE_N, oSRAM_OE_N
);

// Clocks
input   wire            iCLK50;
input   wire            iCLK27;
// User Data
input   wire            iReset;
input   wire            iValid;
input   wire    [15:0]  iIns;
output  wire            oComplete;
output  wire            oVGA_CTRL_CLK;
output  wire            oAUDCtrlClk;
output  wire            oDLY_RST;
// VGA Interface
output  wire            oVGA_CLK;
output  wire            oVGA_HS;
output  wire            oVGA_VS;
output  wire            oVGA_BLANK;
output  wire            oVGA_SYNC;
output  wire    [9:0]   oVGA_R;
output  wire    [9:0]   oVGA_G;
output  wire    [9:0]   oVGA_B;
// Bypress 27 MHz
output wire oTD_RESET;
// SRAM Interface
inout	wire    [15:0]	ioSRAM_DQ;		//	SRAM Data bus 16 Bits
output	wire    [17:0]	oSRAM_ADDR;		//	SRAM Address bus 18 Bits
output  wire	        oSRAM_UB_N;		//	SRAM High-byte Data Mask
output	wire	        oSRAM_LB_N;		//	SRAM Low-byte Data Mask
output	wire	        oSRAM_WE_N;		//	SRAM Write Enable
output	wire	        oSRAM_CE_N;		//	SRAM Chip Enable
output	wire	        oSRAM_OE_N;		//	SRAM Output Enable

// Internal Data Registers
reg     [9:0]   rX1;
reg     [9:0]   rY1;
reg     [9:0]   rX2;
reg     [9:0]   rY2;
reg     [7:0]   rColor;
reg     [7:0]   rChar;

// Helper States
wire    [9:0]   wVGA_X, wVGA_Y;
wire    [9:0]   muxVGA_X, muxVGA_Y;
wire    [9:0]   wVGA_R, wVGA_G, wVGA_B;
wire            wVGA_CTRL_CLK;
wire    [19:0]  wVGA_ADDR;

assign oVGA_CTRL_CLK = wVGA_CTRL_CLK;

assign wVGA_R = {wSBOut[7:6], 8'd0};
assign wVGA_G = {wSBOut[5:4], 8'd0};
assign wVGA_B = {wSBOut[3:2], 8'd0};

// During sync pulses we drive coordinates
// with the write coordinates, during non-sync
// we drive the SRAM for read coordinates.
assign muxVGA_X = (wWE_N) ? wVGA_X : wXCoord;
assign muxVGA_Y = (wWE_N) ? wVGA_Y : wYCoord;

wire    [15:0]  wSBOut;
wire            wSync = (~oVGA_HS | ~oVGA_VS);

reg             rValidChar = 0;
reg             rValidLine = 0;
reg             rValidPixel = 0;

reg             rSetComplete = 0;

wire            wCharComplete;
wire            wLineComplete;
wire            wPixelComplete;

reg             rReset = 0;
reg     [20:0]   rResetCount;
wire            wReset = iReset | rReset;

assign oComplete = wCharComplete  |
                   wLineComplete  |
                   wPixelComplete |
                   rSetComplete;

wire            wWE_N;
wire            wLine_WE_N;
wire            wChar_WE_N;
wire            wPixel_WE_N;
// This kind of modularity will allow us to add new modules later
// without having to worry too much about contention of the lines.
// Since we can depend on rValid to be high until the equivalent
// complete.
assign wWE_N = (rValidLine) ? wLine_WE_N :
               (rValidPixel) ? wPixel_WE_N :
               (rValidChar) ? wChar_WE_N : 1'b1;

wire    [9:0]   wXCoord, wYCoord;
wire    [9:0]   wLineXCoord, wLineYCoord;
wire    [9:0]   wCharXCoord, wCharYCoord;
wire    [9:0]   wPixelXCoord, wPixelYCoord;

assign wXCoord = (rValidLine) ? wLineXCoord :
                 (rValidPixel) ? wPixelXCoord :
                 (rValidChar) ? wCharXCoord : 10'bzzz;
assign wYCoord = (rValidLine) ? wLineYCoord :
                 (rValidPixel) ? wPixelYCoord :
                 (rValidChar) ? wCharYCoord : 10'bzzz;

wire    [1:0]   wR, wG, wB;
wire    [1:0]   wLine_R, wLine_G, wLine_B;
wire    [1:0]   wChar_R, wChar_G, wChar_B;
wire    [1:0]   wPixel_R, wPixel_G, wPixel_B;

assign wR = (rValidLine) ? wLine_R :
            (rValidPixel) ? wPixel_R :
            (rValidChar) ? wChar_R : 2'bzz;
assign wG = (rValidLine) ? wLine_G :
            (rValidPixel) ? wPixel_G :
            (rValidChar) ? wChar_G : 2'bzz;
assign wB = (rValidLine) ? wLine_B :
            (rValidPixel) ? wPixel_B :
            (rValidChar) ? wChar_B : 2'bzz;

SRAM_SCREEN_BUFFER sram_screen_buffer
 (
     .iCLK(wVGA_CTRL_CLK), .iClear(wReset),
     // User Data
     .iCoordX(muxVGA_X), .iCoordY(muxVGA_Y),
     .iWE_N(wWE_N),
     .iWriteR(wR), .iWriteG(wG), .iWriteB(wB),
     .oReadData(wSBOut),
     // Actual SRAM
     .ioSRAM_DQ(ioSRAM_DQ), .oSRAM_ADDR(oSRAM_ADDR),
     .oSRAM_UB_N(oSRAM_UB_N), .oSRAM_LB_N(oSRAM_LB_N),
     .oSRAM_WE_N(oSRAM_WE_N),
     .oSRAM_CE_N(oSRAM_CE_N), .oSRAM_OE_N(oSRAM_OE_N)
);

Brensenham br(
    .iLock(wSync), .iCLK(wVGA_CTRL_CLK),
    // User Inputs
    .iValidLine(rValidLine),
    .iX1(rX1), .iY1(rY1),
    .iX2(rX2), .iY2(rY2),
    .iColor(rColor),
    .oLineComplete(wLineComplete),
    // Screen Buffer Hookup
    .oXCoord(wLineXCoord), .oYCoord(wLineYCoord),
    .oWE_N(wLine_WE_N),
    .oWriteR(wLine_R), .oWriteG(wLine_G), .oWriteB(wLine_B)
);

ASCIIVGA asciivga
(
    .iLock(wSync), .iCLK(wVGA_CTRL_CLK),
    // User Inputs
    .iValidChar(rValidChar),
    .iX1(rX1), .iY1(rY1),
    .iChar(rChar),
    .iColor(rColor),
    // Output to signify completion
    .oCharComplete(wCharComplete),
    // Screen Buffer Hookup
    .oXCoord(wCharXCoord), .oYCoord(wCharYCoord),
    .oWE_N(wChar_WE_N),
    .oWriteR(wChar_R), .oWriteG(wChar_G), .oWriteB(wChar_B)
);

PixelPlot pixelplot
(
    .iLock(wSync), .iCLK(wVGA_CTRL_CLK),
    // User Inputs
    .iValidPixel(rValidPixel),
    .iX1(rX1), .iY1(rY1),
    .iColor(rColor),
    // Output to signify completion
    .oPixelComplete(wPixelComplete),
    // Screen Buffer Hookup
    .oXCoord(wPixelXCoord), .oYCoord(wPixelYCoord),
    .oWE_N(wPixel_WE_N),
    .oWriteR(wPixel_R), .oWriteG(wPixel_G), .oWriteB(wPixel_B)
);

VGACNT vgacnt
(
    //Clocks
    .iCLK50(iCLK50), .iCLK27(iCLK27),
    // Host Side
    .oVGA_ADDR(wVGA_ADDR),
    // Data Side
    .iVGA_R(wVGA_R), .iVGA_G(wVGA_G), .iVGA_B(wVGA_B),
    .oVGA_X(wVGA_X), .oVGA_Y(wVGA_Y), .oVGA_CTRL_CLK(wVGA_CTRL_CLK),
    .oAUDCtrlClk(oAUDCtrlClk),
    .oDLY_RST(oDLY_RST),
    // VGA Side
    .oVGA_CLK(oVGA_CLK),
    .oVGA_HS(oVGA_HS), .oVGA_VS(oVGA_VS),
    .oVGA_BLANK(oVGA_BLANK), .oVGA_SYNC(oVGA_SYNC),
    .oVGA_R(oVGA_R), .oVGA_G(oVGA_G), .oVGA_B(oVGA_B),
    // Bypress 27 MHz
    .oTD_RESET(oTD_RESET)
);

// GPU Control Logic
always@(posedge wVGA_CTRL_CLK) begin
    if(iReset)
        begin
            // reset all variables
            rX1 <= 0;
            rY1 <= 0;
            rX2 <= 0;
            rY2 <= 0;
            rColor <= 0;
            rChar <= 0;
            rValidChar <= 0;
            rValidPixel <= 0;
            rValidLine <= 0;
            rSetComplete <= 0;
        end
    /*else if(rValidChar)   // This may be un needed
        begin
            if(wCharComplete)
                begin
                    rValidChar <= 0;
                end
        end
    else if(rValidPixel)
        begin
            if(wPixelComplete)
                begin
                    rValidPixel <= 0;
                end
        end
    else if(rValidLine)
        begin
            if(wLineComplete)
                begin
                    rValidLine <= 0;
                end
        end*/
    else if(rReset)
        begin
            rResetCount <= rResetCount + 1;
            if(rResetCount[20] == 1)
                begin
                    rSetComplete <= 1;
                end
        end
    else if(~iValid)
        begin
            rReset <= 0;
            rValidChar <= 0;
            rValidPixel <= 0;
            rValidLine <= 0;
            rSetComplete <= 0;
        end
    else
        begin
            case(iIns[`Op])
                // SET
                `setX1:
                    begin
                        rX1 <= iIns[`Data];
                        rSetComplete <= 1;
                    end
                `setY1:
                    begin
                        rY1 <= iIns[`Data];
                        rSetComplete <= 1;
                    end
                `setX2:
                    begin
                        rX2 <= iIns[`Data];
                        rSetComplete <= 1;
                    end
                `setY2:
                    begin
                        rY2 <= iIns[`Data];
                        rSetComplete <= 1;
                    end
                `setChar:
                    begin
                        rChar <= iIns[`Char];
                        rSetComplete <= 1;
                    end
                // Draw
                `drawChar:
                    begin
                        rValidChar <= 1;
                        rColor <= iIns[`Color];
                    end
                `drawLine:
                    begin
                        rValidLine <= 1;
                        rColor <= iIns[`Color];
                    end
                `drawPixel:
                    begin
                        rValidPixel <= 1;
                        rColor <= iIns[`Color];
                    end
                `clear:
                    begin
                        rReset <= 1;
                        rResetCount <= 0;
                        rSetComplete <= 1;
                    end
            endcase
        end

end // end posedge iCLK

endmodule
